# Copyright (C) 2012, Aleksey Lim
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import types
from gettext import gettext as _

import active_document as ad
enforce = ad.util.enforce
util = ad.util

from restful_document import env


def restful_method(**kwargs):

    def decorate(func):
        func.is_restful_method = True
        func.restful_cls_kwargs = kwargs
        return func

    return decorate


class Metadata(object):

    def __init__(self, classes):
        self._methods = [{}, {}, {}]

        for meth in _list_methods(classes):
            methods = self._methods[meth.scope].setdefault(meth.document, {})
            if meth.cmd:
                enforce(meth.cmd not in methods,
                        _('Method "%s" already exists in "%s"'),
                        meth.cmd, methods.get(meth.cmd))
                methods[meth.cmd] = meth
            else:
                enforce(meth.method not in methods,
                        _('"%s" method already exists in "%s"'),
                        meth.method, methods.get(meth.method))
                methods[meth.method] = meth

    def get_method(self):
        enforce(len(env.request.path) <= 3, env.BadRequest,
                _('Requested path consists of more than three parts'))
        if len(env.request.path) == 3:
            env.request.query['prop'] = env.request.path.pop()

        scope = len(env.request.path)
        if scope == 0:
            document = None
        else:
            document = env.request.path[0]
            enforce(document in self._methods[scope], env.BadRequest,
                    _('Unknown document type, "%s"'), document)

        if 'cmd' in env.request.query:
            method_name = env.request.query.pop('cmd')
        else:
            method_name = env.request.method

        return self._methods[scope][document].get(method_name)


class Method(object):

    def __init__(self, cls, scope, cb,
            method, cmd=None, mime_type='application/json'):
        self.cls = cls
        self.document = cls.__name__.lower() if scope else None
        self.scope = scope
        self.method = method
        self.cmd = cmd
        self.mime_type = mime_type
        self._cb = cb

    def __str__(self):
        return str(self._cb)

    def __call__(self):
        try:
            result = self._call()
        except TypeError:
            util.exception()
            raise env.BadRequest(_('Inappropriate arguments'))
        env.responce.setdefault('Content-Type', self.mime_type)
        return result

    def _call(self):
        return self._cb(**env.request.query)


class _ObjectMethod(Method):

    def _call(self):
        guid = env.request.path[1]
        doc = self.cls(guid)
        return self._cb(doc, **env.request.query)


def _list_methods(classes):
    for cls in classes:
        for attr in [getattr(cls, i) for i in dir(cls)]:
            if not hasattr(attr, 'is_restful_method'):
                continue
            method_cls = Method
            if isinstance(attr, types.FunctionType):
                slot = 0
            elif isinstance(attr, types.MethodType):
                if attr.im_self is not None:
                    slot = 1
                else:
                    method_cls = _ObjectMethod
                    slot = 2
            else:
                raise RuntimeError(_('Incorrect RESTful method for %r') % attr)
            yield method_cls(cls, slot, attr, **attr.restful_cls_kwargs)
