# Copyright (C) 2012, Aleksey Lim
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import signal
import atexit
import logging
from optparse import OptionParser
from os.path import basename, join, abspath, exists
from gettext import gettext as _

from gevent.wsgi import WSGIServer

import active_document as ad
import restful_document as rd
enforce = ad.util.enforce
util = ad.util

from restful_document import env, printf


class Server(object):

    def __init__(self, name, description, version, homepage):
        self._classes = []
        self._name = name
        self._description = description
        self._version = version
        self._homepage = homepage
        self._ssl_args = {}

        self._parser = OptionParser(
                usage='%prog [OPTIONS] [COMMAND]',
                description=self._description,
                add_help_option=False)
        self._parser.print_version = \
                lambda: sys.stdout.write('%s\n' % self._version)
        self._parser.add_option('-h', '--help',
                help=_('show this help message and exit'),
                action='store_true')
        self._parser.add_option('-V', '--version',
                help=_('show version number and exit'),
                action='version')

        util.Option.seek('active-document', ad.env)
        util.Option.seek('main', env)

        util.Option.bind(self._parser, [
                '/etc/%s.conf' % self._name,
                '~/.config/%s/config' % self._name,
                ])
        self._options, self._args = self._parser.parse_args()
        util.Option.merge(self._options)

    def serve_forever(self, classes, **ssl_args):
        self._classes = classes
        if [i for i in ssl_args.values() if i is not None]:
            self._ssl_args = ssl_args

        if not self._args and not self._options.help:
            prog = basename(sys.argv[0])
            print 'Usage: %s [OPTIONS] [COMMAND]' % prog
            print '       %s -h|--help' % prog
            print
            print self._parser.description
            print _HELP % self._homepage
            exit(0)

        if self._options.help:
            self._parser.print_help()
            print _HELP % self._homepage
            exit(0)

        command = self._args.pop(0)

        if not env.debug.value:
            logging_level = logging.WARNING
        elif env.debug.value == 1:
            logging_level = logging.INFO
        else:
            logging_level = logging.DEBUG
        logging_format = '%(asctime)s %(levelname)s %(name)s: %(message)s'
        if env.foreground.value or command not in ['start']:
            logging_format = '-- %s' % logging_format
        logging.basicConfig(level=logging_level, format=logging_format)

        try:
            enforce(hasattr(self, '_cmd_' + command),
                    _('Unknown command "%s"') % command)
            exit(getattr(self, '_cmd_' + command)() or 0)
        except Exception:
            printf.exception(_('Aborted %s due to error'), self._name)
            exit(1)
        finally:
            printf.flush_hints()

    def _cmd_config(self):
        if self._args:
            opt = self._args.pop(0)
            enforce(opt in util.Option.items, _('Unknown option "%s"'), opt)
            exit(0 if bool(util.Option.items[opt].value) else 1)
        else:
            print '\n'.join(util.Option.export())

    def _cmd_start(self):
        pidfile, pid = self._check_for_instance()
        if pid:
            logging.warning(_('%s is already run with pid %s'),
                    self._name, pid)
            return 1
        if env.foreground.value:
            self._launch()
        else:
            if not exists(env.logdir.value):
                os.makedirs(env.logdir.value)
            enforce(os.access(env.logdir.value, os.W_OK),
                    _('No write access to %s'), env.logdir.value)
            if not exists(env.rundir.value):
                os.makedirs(env.rundir.value)
            enforce(os.access(env.rundir.value, os.W_OK),
                    _('No write access to %s'), env.rundir.value)
            self._forward_stdout()
            self._daemonize(pidfile)
        return 0

    def _cmd_stop(self):
        __, pid = self._check_for_instance()
        if pid:
            os.kill(pid, signal.SIGTERM)
            return 0
        else:
            logging.warning(_('%s is not run'), self._name)
            return 1

    def _cmd_status(self):
        __, pid = self._check_for_instance()
        if pid:
            printf.info(_('%s started'), self._name)
            return 0
        else:
            printf.info(_('%s stopped'), self._name)
            return 1

    def _cmd_reload(self):
        __, pid = self._check_for_instance()
        if pid:
            os.kill(pid, signal.SIGHUP)
            logging.info(_('Reload %s process'), self._name)

    def _launch(self):
        logging.info(_('Start %s on %s:%s'),
                self._name, env.host.value, env.port.value)

        httpd = WSGIServer((env.host.value, env.port.value),
                rd.Router(self._classes), **self._ssl_args)

        def sigterm_cb(signum, frame):
            logging.info(_('Got signal %s, will stop %s'), signum, self._name)
            httpd.stop()

        def sighup_cb(signum, frame):
            logging.info(_('Reload %s on SIGHUP signal'), self._name)
            self._forward_stdout()

        signal.signal(signal.SIGINT, sigterm_cb)
        signal.signal(signal.SIGTERM, sigterm_cb)
        signal.signal(signal.SIGHUP, sighup_cb)

        try:
            httpd.serve_forever()
        finally:
            httpd.stop()
            ad.close()

    def _check_for_instance(self):
        pid = None
        pidfile = join(env.rundir.value, '%s.pid' % self._name)
        if exists(pidfile):
            try:
                pid = int(file(pidfile).read().strip())
                os.getpgid(pid)
            except (ValueError, OSError):
                pid = None
        return pidfile, pid

    def _forward_stdout(self):
        log_path = abspath(join(env.logdir.value, '%s.log' % self._name))
        sys.stdout.flush()
        sys.stderr.flush()
        logfile = file(log_path, 'a+')
        os.dup2(logfile.fileno(), sys.stdout.fileno())
        os.dup2(logfile.fileno(), sys.stderr.fileno())
        logfile.close()

    def _daemonize(self, pid_path):
        pid_path = abspath(pid_path)

        if os.fork() > 0:
            # Exit parent of the first child
            return

        # Decouple from parent environment
        os.chdir(os.sep)
        os.setsid()

        if os.fork() > 0:
            # Exit from second parent
            # pylint: disable-msg=W0212
            os._exit(0)

        # Redirect standard file descriptors
        if not sys.stdin.closed:
            stdin = file('/dev/null')
            os.dup2(stdin.fileno(), sys.stdin.fileno())

        pidfile = file(pid_path, 'w')
        pidfile.write(str(os.getpid()))
        pidfile.close()
        atexit.register(lambda: os.remove(pid_path))

        try:
            self._launch()
            status = 0
        except Exception:
            logging.exception(_('Abort sugar-server due to error'))
            status = 1

        exit(status)


_HELP = """
Commands:
  start                 start in daemon mode
  stop                  stop daemon
  status                check for launched daemon
  reload                reopen log files in daemon mode
  config                output current configuration

See %s."""
