# Copyright (C) 2012, Aleksey Lim
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import hashlib
import logging
from gettext import gettext as _

from M2Crypto import DSA

import active_document as ad
enforce = ad.util.enforce
util = ad.util

from restful_document import env
from restful_document.metadata import restful_method
from restful_document.document import Document


_logger = logging.getLogger('sugar_stats')


class User(Document):

    @ad.active_property(slot=1, prefix='N', full_text=True)
    def nickname(self, value):
        return value

    @ad.active_property(ad.StoredProperty)
    def color(self, value):
        return value

    @ad.active_property(slot=2, prefix='S', permissions=ad.ACCESS_CREATE)
    def machine_sn(self, value):
        return value

    @ad.active_property(slot=3, prefix='U', permissions=ad.ACCESS_CREATE)
    def machine_uuid(self, value):
        return value

    @ad.active_property(ad.StoredProperty, permissions=ad.ACCESS_CREATE)
    def pubkey(self, value):
        return value

    @classmethod
    @restful_method(method='POST')
    def restful_post(cls):
        props = env.request.content
        enforce('pubkey' in props,
                _('Property "pubkey" is required for user registeration'))
        props['manual_guid'], props['pubkey'] = \
                _load_pubkey(props['pubkey'].strip())
        doc = cls.create(props)
        return {'guid': doc.guid}

    @classmethod
    def verify(cls, guid, signature):
        try:
            pubkey = cls(guid, raw=True)['pubkey']
        except ad.NotFound:
            raise env.Unauthorized(_('Principal user does not exist'))
        if env.trust_users.value:
            return
        pubkey_path = util.TempFilePath(text=pubkey)
        pubkey = DSA.load_pub_key(pubkey_path)
        data = hashlib.sha1(guid).digest()
        enforce(pubkey.verify_asn1(data, signature.decode('hex')),
                env.Forbidden, _('Wrong principal credentials'))

    @classmethod
    def init(cls, final_cls=None):
        # `restful_document.User` is the final class
        # for all possible user documents
        Document.init(User)

    def on_create(self, properties):
        properties['guid'] = properties.pop('manual_guid')
        Document.on_create(self, properties)


def _load_pubkey(pubkey):
    try:
        src_path = util.TempFilePath(text=pubkey)
        # SSH key needs to be converted to PKCS8 to ket M2Crypto read it
        pubkey_pkcs8 = util.assert_call(
                ['ssh-keygen', '-f', src_path, '-e', '-m', 'PKCS8'])
    except Exception:
        message = _('Cannot read DSS public key gotten for registeration')
        util.exception(message)
        if env.trust_users.value:
            _logger.warning(_('Failed to read registration pubkey, ' \
                    'but we trust users'))
            # Keep SSH key for further converting to PKCS8
            pubkey_pkcs8 = pubkey
        else:
            raise env.Forbidden(message)

    return str(hashlib.sha1(pubkey.split()[1]).hexdigest()), pubkey_pkcs8
